package ml.active;

import java.io.*;
import java.util.*;

import weka.core.*;

public class XActiveEvaluation {
    private static XActiveEvaluation[] m_lastRunResult = null;
    private static String m_lastOutFileName = null;
    private static boolean m_doNotWriteToFile = false;
    /** The weight of all incorrectly classified instances. */
    private double m_Incorrect;

    /** The weight of all correctly classified instances. */
    private double m_Correct;

    /** The weight of all unclassified instances. */
    private double m_Unclassified;

    /*** The weight of all instances that had no class assigned to them. */
    private double m_MissingClass;

    /** The weight of all instances that had a class assigned to them. */
    private double m_WithClass;

    /** Random instance for dataset shuffling and resampling */
    private Random m_Random = null;

    private FastVector m_pctCorrectVec = null;
    private double m_foldCorrect;
    private double m_foldWithClass;

    private long roundStartTime, roundEndTime;
    private FastVector roundTimeSec = new FastVector();

    private static FastVector m_foldTestClassHisto = new FastVector();


    /**
     * Initializes all the counters for the evaluation.
     *
     * @param data set of training instances, to get some header
     * information and prior class distribution information
     * @exception Exception if the class is not defined
     */
    public XActiveEvaluation(Instances data) throws Exception {
        if (!data.classAttribute().isNominal()) {
            throw new Exception(
                    "XActiveEvaluation: Cannot handle numeric classes!");
        }
        this.m_pctCorrectVec = new FastVector();


        this.m_Incorrect = 0;
        this.m_Correct = 0;
        this.m_Unclassified = 0;
        this.m_MissingClass = 0;
        this.m_WithClass = 0;
        this.m_foldCorrect = 0;
        this.m_foldWithClass = 0;
        //{{INIT_CONTROLS
        //}}
    }

    public static XActiveEvaluation[] getLastRunResult() {
        return m_lastRunResult;
    }

    public static String getLastOutFileName() {
        return m_lastOutFileName;
    }

    public static void setDoNotWriteToFile(boolean val) {
        m_doNotWriteToFile = val;
    }

    /********** total counters **********/
    public final double numInstances() {
        return m_WithClass;
    }

    public final double incorrect() {
        return m_Incorrect;
    }

    public final double pctIncorrect() {
        return 100 * m_Incorrect / m_WithClass;
    }

    public final double correct() {
        return m_Correct;
    }

    public final double pctCorrect() {
        return 100 * m_Correct / m_WithClass;
    }

    public final double unclassified() {
        return m_Unclassified;
    }

    public final double pctUnclassified() {
        return 100 * m_Unclassified / m_WithClass;
    }

    public final double errorRate() {
        return m_Incorrect / m_WithClass;
    }

    /************** fold counters *************/
    public double foldPctCorrect() {
        return 100 * this.m_foldCorrect / this.m_foldWithClass;
    }

    public void resetFoldCounters() {
        this.m_foldCorrect = 0;
        this.m_foldWithClass = 0;
    }

    public void saveFoldCounters() {
        double pct = this.foldPctCorrect();
        this.m_pctCorrectVec.addElement(new Double(pct));
        this.roundTimeSec.addElement(new Double((roundEndTime-roundStartTime) / 1000.0));
    }

    public FastVector getPctCorrectVec() {
        return this.m_pctCorrectVec;
    }

    public FastVector getRoundTimeSec(){
        return roundTimeSec;
    }

    private static String makeOptionString(XActiveLearner learner) {

        StringBuffer optionsText = new StringBuffer("");

        // General options
        optionsText.append("\n\nGeneral options:\n\n");
        optionsText.append("-t <name of training file>\n");
        optionsText.append("\tSets training file.\n");
        optionsText.append("-T <name of test file>\n");
        optionsText.append("\tSets test file. If missing, a cross-validation\n");
        optionsText.append("\twill be performed on the training data.\n");
        optionsText.append("-i <interval size>\n");
        optionsText.append(
                "\tSets interval size for learning curve (default: 1).\n");
        optionsText.append("-m <interval size>\n");
        optionsText.append("\tSets maximum number of instances for learning curve (default: size of training set).\n");
        optionsText.append("-c <class index>\n");
        optionsText.append("\tSets index of class attribute (default: last).\n");
        optionsText.append("-x <number of folds>\n");
        optionsText.append(
                "\tSets number of folds for cross-validation (default: 10).\n");
        optionsText.append("-s <random number seed>\n");
        optionsText.append(
                "\tSets random number seed for cross-validation (default: 1).\n");
        optionsText.append("-o <name of output file>\n");
        optionsText.append("\tName of the file to which the result is written (deault: activeResult.txt).\n");

        // Get scheme-specific options
        if (learner instanceof OptionHandler) {
            optionsText.append("\nOptions specific to "
                               + learner.getClass().getName()
                               + ":\n\n");
            Enumeration enm = ((OptionHandler) learner).listOptions();
            while (enm.hasMoreElements()) {
                Option option = (Option) enm.nextElement();
                optionsText.append(option.synopsis() + '\n');
                optionsText.append(option.description() + "\n");
            }
        }
        return optionsText.toString();
    }

    public void evaluateModelOnce(XActiveLearner learner,
                                  InstanceContainer instance) throws Exception {
        updateStatsForClassifier(learner.classifyInstance(instance), instance);
    }

    public void evaluateModel(XActiveLearner learner, DataContainer data) throws
            Exception {
        int i, size = data.size();
        System.out.print("Testing");
        for (i = 0; i < size; i++) {
            evaluateModelOnce((XActiveLearner) learner, data.getInstance(i));
        }
        System.out.println("done.");
    }

    public static void crossValidateModel(XActiveEvaluation[] testingEvaluation,
                                          XActiveLearner learner,
                                          DataRealizator data,
                                          int numFolds,
                                          int maxNumInst) throws Exception {
        int poolSize;
        int i, j, toQuery;
        DataContainer pool, test, currTrain;
        Instances testAsInstances;

        /* startify the data according to number of folds.
         * the propotion of the labels in each "fold cluster" corresponding the
         * label's  propotion of the full data.
         * (Note that the writer assumed that the data is randomized)
         */
        data.stratify(numFolds);

        // Do the folds
        for (i = 0; i < numFolds; i++) {
            // start with an empty training set and a full pool
            /*ron learner.startActiveLearningSession(); */

            /* I think that the following line produce a (conceptual) bug!
             * tmp contains an 0.6 (random) instances. hence, the fold index
             * is irrelevant as well the startifation of the data (see, former code lines)
             *
             * also note that there might be an overlap between different
             * cv-folds (the implementation randomly choose instances for
             * each fold)
             */
            Instances poolInInstancesFormat = data.trainCVRatio(numFolds, i,
                    0.6);
            poolSize = poolInInstancesFormat.size();
            pool = learner.create(poolInInstancesFormat);
            /*testCVRatio must be called after trainCVRatio !! */
            testAsInstances = data.testCVRatio(numFolds, i, 0.6);
            setFoldTestClassHistogram(testAsInstances);
            test = learner.create(testAsInstances, pool);

            System.out.println("\nTest: " + test.toString());

            currTrain = learner.create(initiateTraining(poolInInstancesFormat),
                                       pool);

            learner.init(currTrain);

            /*ron learner.setDataAndPool(tmp, pool); */

            if (maxNumInst > poolSize) {
                maxNumInst = poolSize;
            }

            /* ?? */
            if (testingEvaluation[maxNumInst - 1] == null) {
                testingEvaluation[maxNumInst - 1] =
                        new XActiveEvaluation(data.emptySet());
            }

            for (j = 0; j < maxNumInst; j++) {
                testingEvaluation[j].roundStartTime = System.currentTimeMillis();

                // query and update training set and pool
                System.out.print("Getting instance to query.");
                toQuery = learner.instanceToQuery(pool);
                System.out.println("done.");
                // the learner is allowed to erase instances from the currTrain
                currTrain.addInstance(pool.getInstance(toQuery));
                pool.deleteInstance(toQuery);
                // build classifier
                learner.buildClassifier(currTrain);

                if (testingEvaluation[j] != null) {
                    testingEvaluation[j].resetFoldCounters();
                    /* results are updated both in the overall counters and in the
                       fold counters */
                    testingEvaluation[j].evaluateModel((XActiveLearner) learner,
                            test);
                    testingEvaluation[j].roundEndTime = System.currentTimeMillis();
                    testingEvaluation[j].saveFoldCounters();
                }
                System.out.println("checkpoint run:fold:instances = "
                                   + XActiveEvaluationPlotter.m_currRun + ":" +
                                   i + ":" + j + " "
                                   +
                                   testingEvaluation[j].
                                   outputFoldSuccessPercentage());
            }
        }
    }

    /**
     * setTrivialSuccessStat
     *
     * @param test DataContainer
     */
    private static void setFoldTestClassHistogram(Instances test) throws
            Exception {
        m_foldTestClassHisto.addElement(test.classHistorgram());
    }

    private void updateStatsForClassifier(double prediction,
                                          InstanceContainer instance) throws
            Exception {
        if (!Instance.isMissingValue(instance.classValue())) {
            int actualClass = (int) instance.classValue();
            m_WithClass += instance.weight();
            this.m_foldWithClass += instance.weight();
            if (Instance.isMissingValue(prediction)) {
                m_Unclassified += instance.weight();
                return;
            }
            int predictedClass = (int) prediction;
            if (predictedClass != actualClass) {
                m_Incorrect += instance.weight();
            } else {
                m_Correct += instance.weight();
                this.m_foldCorrect += instance.weight();
            }
        } else {
            m_MissingClass += instance.weight();
        }
    }

    public String toSummaryString(String title) {
        double mae, mad = 0;
        StringBuffer text = new StringBuffer();

        text.append(title + "\n");
        try {
            if (m_WithClass > 0) {
                text.append("Correctly Classified Instances     ");
                text.append(Utils.doubleToString(correct(), 12, 4) + "     " +
                            Utils.doubleToString(pctCorrect(),
                                                 12, 4) + " %\n");
                text.append("Incorrectly Classified Instances   ");
                text.append(Utils.doubleToString(incorrect(), 12, 4) + "     " +
                            Utils.doubleToString(pctIncorrect(),
                                                 12, 4) + " %\n");
            }
            if (Utils.gr(unclassified(), 0)) {
                text.append("UnClassified Instances             ");
                text.append(Utils.doubleToString(unclassified(), 12, 4) +
                            "     " +
                            Utils.doubleToString(pctUnclassified(),
                                                 12, 4) + " %\n");
            }
            text.append("Total Number of Instances          ");
            text.append(Utils.doubleToString(m_WithClass, 12, 4));
            if (m_MissingClass > 0) {
                text.append("\nIgnored Class Unknown Instances            ");
                text.append(Utils.doubleToString(m_MissingClass, 12, 4));
            }
        } catch (Exception ex) {
            // Should never occur since the class is known to be nominal
            // here
            System.err.println("Arggh - Must be a bug in Evaluation class");
        }
        return text.toString();
    }

    public String outputFoldSuccessPercentage() {
        return " " + Utils.doubleToString(this.foldPctCorrect(), 12, 4) + "\n";
    }

    public String outputSuccessPercentage(int numOfInst) {
        return "" + numOfInst + " " + Utils.doubleToString(pctCorrect(), 12, 4) +
                "\n";
    }

    /**
     * Evaluates a classifier with the options given in an array of
     * strings. <p>
     *
     * Valid options are: <p>
     *
     * -t name of training file <br>
     * Name of the file with the training data. (required) <p>
     *
     * -T name of test file <br>
     * Name of the file with the test data. If missing a cross-validation
     * is performed. <p>
     *
     * -c class index <br>
     * Index of the class attribute (1, 2, ...; default: last). <p>
     *
     * -i interval size <br>
     * <p>
     *
     * -m maximal checkpoint <br>
     * <p>
     *
     * -x number of folds <br>
     * The number of folds for the cross-validation (default: 10). <p>
     *
     * -y number of folds in random pick <br>
     * The number of folds in random pick (default: 10). <p>
     *
     * -s random number seed <br>
     * Random number seed for the cross-validation (default: 1). <p>
     *
     * -o name of output file <br>
     * Name of the file to which the result is written (deault: activeResult.txt) <p>
     *
     * @param classifier machine learning classifier
     * @param options the array of string containing the options
     * @exception Exception if model could not be evaluated successfully
     * @return a string describing the results */
    public static String evaluateModel(XActiveLearner learner, String[] options) throws
            Exception {
        DataRealizator train = null;
        Instances test = null, testWithoutStrings = null;
        int seed = 1, folds = 10, rndInst = -1, classIndex = -1, interval = 1,
                max;
        int i;
        String trainFileName, testFileName, classIndexString, outFileName,
                seedString, foldsString, rndInstString, intervalString,
                maxString;
        XActiveEvaluation[] testingEvaluation;
        StringBuffer text = new StringBuffer();
        BufferedReader trainReader = null, testReader = null;
        Random random;
        StringBuffer schemeOptionsText = null;
        FileOutputStream outFile;

        /*** Initiating options and data ***/

        try {
            // Get basic options (options the same for all schemes)
            classIndexString = Utils.getOption('c', options);
            if (classIndexString.length() != 0) {
                classIndex = Integer.parseInt(classIndexString);
            }
            trainFileName = Utils.getOption('t', options);
            testFileName = Utils.getOption('T', options);
            if (trainFileName.length() == 0) {
                throw new Exception("No training file given.");
            }
            try {
                File file = new File(trainFileName);
                if (file.isDirectory()) {
                    train = new InstancesAscSet(trainFileName);
                } else if (trainFileName.endsWith(".arff")) {
                    trainReader = new BufferedReader(new FileReader(
                            trainFileName));
                    train = new Instances(trainReader);
                    trainReader.close();
                } else {
                    // other format support!!
                    train = new Instances(trainFileName);
                }
                if (testFileName.length() != 0) {
                    if (!(train instanceof Instances)) {
                        throw new Exception(
                                "XActiveEvaluation: InstancesAscSet is not supported with additional test file");
                    }
                    testReader = new BufferedReader(new FileReader(testFileName));
                }
            } catch (Exception e) {
                throw new Exception("Can't open file " + e.getMessage() + '.');
            }
            if (classIndex != -1) {
                if (train instanceof Instances) {
                    ((Instances) train).setClassIndex(classIndex - 1);
                } else {
                    throw new Exception(
                            "InstancesAscSet do not support classIndex");
                }
            } else {
                if (train instanceof Instances) {
                    ((Instances) train).setClassIndex(train.numAttributes() - 1);
                }
            }
            if (classIndex > train.numAttributes()) {
                throw new Exception("Index of class attribute too large.");
            }

            if (testFileName.length() != 0) {
                test = new Instances(testReader);
                if (classIndex != -1) {
                    test.setClassIndex(classIndex - 1);
                } else {
                    test.setClassIndex(test.numAttributes() - 1);
                }
                if (classIndex > test.numAttributes()) {
                    throw new Exception("Index of class attribute too large.");
                }
                testWithoutStrings = new Instances(test);
                testReader.close();
            }
            seedString = Utils.getOption('s', options);
            if (seedString.length() != 0) {
                seed = Integer.parseInt(seedString);
            }
            foldsString = Utils.getOption('x', options);
            if (foldsString.length() != 0) {
                folds = Integer.parseInt(foldsString);
            } else {
                folds = train.defaultNumFolds();
            }
            rndInstString = Utils.getOption('y', options);
            if (rndInstString.length() != 0) {
                rndInst = Integer.parseInt(rndInstString);
            }
            intervalString = Utils.getOption('i', options);
            if (intervalString.length() != 0) {
                interval = Integer.parseInt(intervalString);
            }
            maxString = Utils.getOption('m', options);
            if (maxString.length() != 0) {
                max = Integer.parseInt(maxString);
                if (max > train.numInstances()) {
                    max = train.numInstances();
                }
            } else {
                max = train.numInstances();
            }
            outFileName = Utils.getOption('o', options);
            if (outFileName.length() == 0) {
                outFileName = new String("activeResult");
            }
            // Set options for learner
            if (learner instanceof OptionHandler) {
                for (i = 0; i < options.length; i++) {
                    if (options[i].length() != 0) {
                        if (schemeOptionsText == null) {
                            schemeOptionsText = new StringBuffer();
                        }
                        if (options[i].indexOf(' ') != -1) {
                            schemeOptionsText.append('"' + options[i] + "\" ");
                        } else {
                            schemeOptionsText.append(options[i] + " ");
                        }
                    }
                }
                ((OptionHandler) learner).setOptions(options);
            }
            Utils.checkForRemainingOptions(options);
        } catch (Exception e) {
            throw new Exception("\nWeka exception: " + e.getMessage()
                                + makeOptionString(learner));
        }
        text.append(learner.toString());
        text.append("Database: " + trainFileName +
                    "\n--------------------------\n");
        // Compute proper error estimates
        int toQuery;
        // Setup up evaluation objects
        testingEvaluation = new XActiveEvaluation[max];
        for (i = 0; i < max; i++) {
            if ((i % interval) == 0) {
                testingEvaluation[i] = new XActiveEvaluation(train.emptySet());
            } else {
                testingEvaluation[i] = null;
            }
        }

        /**** Starting Evaluation ***********/

        if (testFileName.length() != 0) {
            Instances tr = (Instances) train;
            if (testingEvaluation[max - 1] == null) {
                testingEvaluation[max -
                        1] = new XActiveEvaluation(train.emptySet());
            }
            text.append("Testing on test file " + testFileName + "\n\n");
            // start with an empty training set and a full pool
            learner.startActiveLearningSession();
            Instances currTrain = tr.emptySet();
            System.out.print("copying data for pool...");
            Instances pool = new Instances(tr);
            System.out.println("done.");
            for (i = 0; i < max; i++) {
                // query and update training set and pool
                toQuery = learner.instanceToQuery(pool);
                currTrain.addNoCopy(pool.instance(toQuery));
                pool.delete(toQuery);
                // build classifier
                learner.buildClassifier(currTrain);
                if (testingEvaluation[i] != null) {
                    testingEvaluation[i].resetFoldCounters();
                    testingEvaluation[i].evaluateModel((XActiveLearner) learner,
                            testWithoutStrings);
                    testingEvaluation[i].saveFoldCounters();
                    text.append(testingEvaluation[i].outputSuccessPercentage(i));
                }
            }
        } else {
            // Testing is via cross-validation on training data
            text.append("Testing with " + folds +
                        "-fold cross validation on training set.\n\n");
            random = new Random(seed);
            random.setSeed(seed);
            System.out.print("Randomizing data...");
            train.randomize(random, folds, rndInst);
            System.out.println("done.");
            crossValidateModel(testingEvaluation,
                               learner,
                               train,
                               folds,
                               max);
            for (i = 0; i < max; i++) {
                if (testingEvaluation[i] != null) {
                    text.append(testingEvaluation[i].outputSuccessPercentage(i));
                }
            }
        }
        if (!m_doNotWriteToFile) {
            outFile = new FileOutputStream(outFileName + ".txt");
            outFile.write(text.toString().getBytes());
            outFile.close();
        }
        m_lastRunResult = testingEvaluation;
        m_lastOutFileName = new String(outFileName);
        return text.toString();
    }


    /**
     * initiates the training with two instances (from each class).
     * the instances are chosen according to thier indices.
     *
     * @param pool Instances
     * @return Instances
     */
    private static Instances initiateTraining(Instances pool) throws Exception {
        Instances training = pool.emptySet();
        InstanceContainer instance = null;
        training.addInstance(pool.getInstance(0));

        for (int i = 0, test = pool.size(); i < test; ++i) {
            instance = pool.getInstance(i);
            if (training.getInstance(0).classValue() == instance.classValue()) {
                continue;
            } else {
                training.addInstance(instance);
                break;
            }
        }
        return training;
    }

    public static FastVector getHistogramByFold() {
        FastVector res = m_foldTestClassHisto;
        m_foldTestClassHisto = new FastVector();
        return res;
    }
}
