package weka.active;

import java.util.*;
import java.io.*;
import weka.active.*;
import weka.core.*;
import weka.clusterers.*;
import weka.classifiers.Classifier;
import weka.classifiers.libsvm.*;

public abstract class SvmActiveLearner extends weka.active.ActiveLearner
                                       implements DifferentFormatHandler
{
    /****** constants ********/
    protected static final int UNCLASSIFIED = -1;

    protected static final int MODE_ACTIVE_NORMAL = 0;
    protected static final int MODE_ACTIVE_NORMAL_CV = 11;
    protected static final int MODE_ACTIVE_CEM = 1;
    protected static final int MODE_ACTIVE_LOO = 2;
    protected static final int MODE_MS_BEST = 3;
    protected static final int MODE_MS_LOO = 4;
    protected static final int MODE_MS_CEM = 5;
    protected static final int MODE_MS_WORST = 6;
    protected static final int MODE_MS_BEST_ALL = 7;
    protected static final int MODE_MS_LOO_ALL = 8;
    protected static final int MODE_MS_CEM_ALL = 9;
    protected static final int MODE_MS_WORST_ALL = 10;
    protected static final int MODE_CEM_PROOF = 12;
    // members
    protected SvmLib m_svm = new SvmLib();
    protected boolean m_classifierExist = false;
    protected int m_firstClassSeen = NO_CLASS;
    private boolean m_waitForUserFlag = false;
    protected Instances m_data;
    protected DataContainer m_pool;
    protected DataContainer m_trainingSet; // a ref to the training set
    protected DataContainer m_fullTrainingSet;
    protected int[] m_wasQueried;
    protected Random m_rndm = null;
    protected int m_seed = 1234567;
    protected FastVector m_holdOut;

    protected static FastVector m_selfTestInfo = null;
    protected static FastVector m_selfTestPosRatio = null;
    protected static FastVector m_selfTestActualPosRatio = null;
    protected static FastVector m_selfTestSVM = null;
    protected static FastVector m_selfTestC45 = null;
    protected static FastVector m_selfTestNN = null;
    protected static FastVector m_selfTestBPM = null;
    protected static FastVector m_selfTestAccC45 = null;
    protected static FastVector m_selfTestAccNN = null;
    protected static FastVector m_selfTestAccBPM = null;
    protected double m_lastEntropyRes;
    protected double m_previousEntropyRes;
    protected double m_lastFullEntropyRes;
    protected double m_previousFullEntropyRes;

    private static final double[] m_gamma = { 0.01, 0.03, 0.05, 0.07, 0.1, 0.3, 0.5, 0.7, 1.0, 2.0, 3.0, 4.0, 5.0 };
    private static final double[] m_degree = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 };
    private static final double[] m_gammaMS = { 0.01, 0.05, 0.1, 0.5, 1.0, 5.0 };
    private static final double[] m_degreeMS = { 1, 4, 7, 10, 13, 16};
    private Random m_rndmForGamma = null;
    private int m_seedForGamma = 84912;
    private double m_sampleRatio = 0.5;
    private static final String m_paramsPath = "./params/";
    private FastVector m_params = null;
    private boolean m_reading = false;
    private int m_paramsIndx = -1;
    private boolean m_polynomial = false; // default is RBF
    private Instances m_origTestSet;
    private DataContainer m_testSet;
    private int m_mode=MODE_ACTIVE_NORMAL;

    private DataContainer convertDataHalfMode(Instances pool) throws Exception
    {
        int i, size;
        DataContainer convertedData=null, sampled=null;
        InstanceContainer ins;
        Instances empty = new Instances(pool, 0);

        // init
        if (this.m_rndmForGamma == null) {
            this.m_rndmForGamma = new Random(m_seedForGamma);
        }
        // create trainingSets
        convertedData = this.m_svm.convertData(pool);
        size = pool.size();
        for (i=0; i<size; i++) {
            empty.addNoCopy(pool.instance(i));
        }
        sampled = sampleForGamma(convertedData, empty, this.m_sampleRatio, this.m_rndmForGamma);
        // now sampled will be used for gamma pick, and convertedData(with empty) will be used for training
        System.out.println("\nTrain:" + convertedData.toString());
        System.out.println("\nSampled:" + sampled.toString());

        if (this.m_params == null) {
            String fileName;
            if (!this.m_polynomial) {
                fileName = new String(m_paramsPath + pool.relationName() + "_cls");
            } else {
                fileName = new String(m_paramsPath + pool.relationName() + "_poly_cls");
            }
            File f = new File(fileName);
            if (f.exists()) {
                FileInputStream inFile = new FileInputStream(f);
                ObjectInputStream inObj = new ObjectInputStream(inFile);
                m_params = (FastVector)(inObj.readObject());
                inObj.close();
                this.m_reading = true;
                this.m_paramsIndx = 0;
            } else {
                m_params = new FastVector();
                this.m_reading = false;
            }
        }

        if (!this.m_polynomial) { // RBF
            if (this.m_reading) {
                this.setGamma(((Double)(this.m_params.elementAt(this.m_paramsIndx))).doubleValue());
                this.m_paramsIndx++;
            } else {
                // adjust parameters
                double suc, bestSuc=-Double.MAX_VALUE;
                int bestGammaIndx=-1;
                for (i=0; i<m_gamma.length; i++) {
                    suc = this.KFoldCV(sampled, 10, m_gamma[i], this.m_rndmForGamma);
                    System.out.println("gamma=" + m_gamma[i] + " suc=" + (suc*100));
                    if (Utils.gr(suc, bestSuc)) {
                        bestSuc = suc;
                        bestGammaIndx = i;
                    }
                }
                this.setGamma(m_gamma[bestGammaIndx]);
                System.out.println("BestGamma=" + m_gamma[bestGammaIndx]);
                this.m_params.addElement(new Double(m_gamma[bestGammaIndx]));
                String fileName = new String(m_paramsPath + pool.relationName() + "_cls");
                File f = new File(fileName);
                FileOutputStream outFile = new FileOutputStream(f);
                ObjectOutputStream outObj = new ObjectOutputStream(outFile);
                outObj.writeObject(this.m_params);
                outObj.close();
            }
        } else {
            if (this.m_reading) {
                this.setDegree(((Double)(this.m_params.elementAt(this.m_paramsIndx))).doubleValue());
                this.m_paramsIndx++;
            } else {
                // adjust parameters
                double suc, bestSuc=-Double.MAX_VALUE;
                int bestDegreeIndx=-1;
                for (i=0; i<m_degree.length; i++) {
                    suc = this.KFoldCV(sampled, 10, m_degree[i], this.m_rndmForGamma);
                    System.out.println("degree=" + m_degree[i] + " suc=" + (suc*100));
                    if (Utils.gr(suc, bestSuc)) {
                        bestSuc = suc;
                        bestDegreeIndx = i;
                    }
                }
                this.setDegree(m_degree[bestDegreeIndx]);
                System.out.println("BestDegree=" + m_degree[bestDegreeIndx]);
                this.m_params.addElement(new Double(m_degree[bestDegreeIndx]));
                String fileName = new String(m_paramsPath + pool.relationName() + "_poly_cls");
                File f = new File(fileName);
                FileOutputStream outFile = new FileOutputStream(f);
                ObjectOutputStream outObj = new ObjectOutputStream(outFile);
                outObj.writeObject(this.m_params);
                outObj.close();
            }
        }
        this.m_pool = convertedData;
        this.m_data = empty;
        return convertedData;
    }

    private DataContainer convertDataFullMode(Instances pool) throws Exception
    {
        DataContainer convertedData = this.m_svm.convertData(pool);
        this.m_pool = convertedData;
        this.m_data = pool;
        return convertedData;
    }

    public DataContainer convertData(Instances pool) throws Exception
    {
        DataContainer data;
        if ((this.m_mode == MODE_ACTIVE_NORMAL) ||
            (this.m_mode == MODE_ACTIVE_NORMAL_CV) ||
            (this.m_mode == MODE_CEM_PROOF))
        {
            data = convertDataHalfMode(pool);
        } else {
            data = convertDataFullMode(pool);
        }
        return data;
    }

    public DataContainer convertData(Instances data, DataContainer former) throws Exception
    {
        if (data.size() > 0) {
            this.m_origTestSet = data;
            this.m_testSet = this.m_svm.convertData(data, former);
            return this.m_testSet;
        }
        return this.m_svm.convertData(data, former);
    }

    public int instanceToQuery(DataContainer pool) throws Exception
    {
        int i, res;
        this.printMeg("");

        res = this.instanceToQueryInitMode(pool);
        if (res != -1) {
            System.out.println("\tquerying in init mode. indx=" + pool.getInstance(res).getIndex());
            return res;
        }
        if ((this.m_mode == MODE_ACTIVE_NORMAL) ||
            (this.m_mode == MODE_ACTIVE_NORMAL_CV) ||
            (this.m_mode == MODE_CEM_PROOF))
        {
            this.m_svm.buildClassifier(this.m_fullTrainingSet);
        }
        return -1;
    }

    public double[] distributionToQuery(DataContainer pool) throws Exception
    {
        return null;
    }

    protected int instanceToQueryRandomMode(DataContainer pool) throws Exception
    {
        int res = (int)(Math.ceil(pool.size() * this.m_rndm.nextDouble()))-1;
        return res;
    }

    public void buildClassifier(DataContainer data) throws Exception
    {
        /* updates the newly added instance (denoted here INS).
         * the implementation assumes that INS was added to the
         * end of data (last index of data); the m_wasQueried attr. is marked according
         * the index of INS.
        */
        this.updateNewQueryResult(data);


        /* m_trainingSet seems to be a reference to the data
        * the code make assumptions about the way the user of this
        * class manipulates that data (for example, adding the queried
        * instance to the end of data before calling buildClassifier)
        */
        if (this.m_trainingSet == null) {
            this.m_trainingSet = data;
        }

        /* m_fullTrainingSet seems to be hold a copy of the data
        *
        */
        if (this.m_fullTrainingSet == null) {
            this.m_fullTrainingSet = this.m_trainingSet.makeEmptyCopy();
            int i, size = this.m_trainingSet.size();
            for (i=0; i<size; i++) {
                this.m_fullTrainingSet.addInstance(this.m_trainingSet.getInstance(i));
            }
        } else {
            this.m_fullTrainingSet.addInstance(this.m_trainingSet.getInstance(this.m_trainingSet.size()-1));
        }

        /* m_selfTestInfo seems to hold self_tests results (for determining hyper params)
        */
        if (m_selfTestInfo.size() < m_fullTrainingSet.size()) {
            if (m_fullTrainingSet.size()-1 != m_selfTestInfo.size()) throw new Exception("missed some queires! selfTestSize=" + m_selfTestInfo.size() + " dataSize=" + m_fullTrainingSet.size());
            m_selfTestInfo.addElement(new FastVector());
        }


        if (m_mode == MODE_CEM_PROOF) {
            if (m_selfTestPosRatio.size() < m_fullTrainingSet.size()) {
                if (m_fullTrainingSet.size()-1 != m_selfTestPosRatio.size()) throw new Exception("missed some queires! selfTestSize=" + m_selfTestPosRatio.size() + " dataSize=" + m_fullTrainingSet.size());
                m_selfTestPosRatio.addElement(new FastVector());
            }
            if (m_selfTestActualPosRatio.size() < m_fullTrainingSet.size()) {
                if (m_fullTrainingSet.size()-1 != m_selfTestActualPosRatio.size()) throw new Exception("missed some queires! selfTestSize=" + m_selfTestActualPosRatio.size() + " dataSize=" + m_fullTrainingSet.size());
                m_selfTestActualPosRatio.addElement(new FastVector());
            }
            if (m_selfTestSVM.size() < m_fullTrainingSet.size()) {
                if (m_fullTrainingSet.size()-1 != m_selfTestSVM.size()) throw new Exception("missed some queires! selfTestSize=" + m_selfTestSVM.size() + " dataSize=" + m_fullTrainingSet.size());
                m_selfTestSVM.addElement(new FastVector());
            }
        }

        /* following two iffs are for dealing with the first time this method
        * is called.
        */

        if (this.m_firstClassSeen == NO_CLASS) {
            // this is the first instance given
            this.m_firstClassSeen = (int)data.getInstance(0).classValue();
        }
        if (!this.m_classifierExist) {
            // until now we have seen only one class
            int i, size = data.size();
            InstanceContainer ins;
            for (i=0; i<size; i++) {
                ins = data.getInstance(i);
                if (((int)ins.classValue()) != this.m_firstClassSeen) {
                    this.m_classifierExist = true;
                    break;
                }
            }
            if (!this.m_classifierExist) {
                this.waitForUser();
                return;
            }
        }

        /* the following code does the actual construction of the classifier.
        */
        if ((this.m_mode == MODE_ACTIVE_NORMAL) ||
            (this.m_mode == MODE_ACTIVE_NORMAL_CV) ||
            (this.m_mode == MODE_CEM_PROOF))
        {
            this.m_svm.buildClassifier(this.m_fullTrainingSet);
            if ((this.m_mode == MODE_ACTIVE_NORMAL) || (this.m_mode == MODE_CEM_PROOF)) {
                this.betterInInfoGain(this.m_fullTrainingSet, this.m_pool);
            } else {
                this.betterInSelfEstimate(this.m_fullTrainingSet, this.m_pool);
            }
            if ((this.m_holdOut.size() <= 2) && Utils.sm(this.m_lastFullEntropyRes, this.m_previousFullEntropyRes)) { // PARAMETER
                this.m_holdOut.addElement(this.m_trainingSet.getInstance(this.m_trainingSet.size()-1));
                this.m_trainingSet.deleteInstance(this.m_trainingSet.size()-1);
                System.out.println("HOLDING. size=" + this.m_holdOut.size());
            } else {
                System.out.println("RETURNING " + this.m_holdOut.size());
                Enumeration instEnum = this.m_holdOut.elements();
                InstanceContainer ins;
                while (instEnum.hasMoreElements()) {
                    ins = (InstanceContainer)(instEnum.nextElement());
                    this.m_trainingSet.addInstance(ins);
                }
                this.m_holdOut.removeAllElements();
            }
            System.out.println("building classifier gamma=" + this.m_svm.getGamma());
            this.m_svm.buildClassifier(this.m_trainingSet);
        } else if ((this.m_mode == MODE_ACTIVE_CEM) || (this.m_mode == MODE_ACTIVE_LOO)) {
            this.paramSelection(this.m_fullTrainingSet);
            System.out.println("building classifier gamma=" + this.m_svm.getGamma());
            this.m_svm.buildClassifier(this.m_fullTrainingSet);
        } else { // MS
            this.modelSelection(this.m_fullTrainingSet);
            this.m_svm.buildClassifier(this.m_fullTrainingSet);
        }
    }

    public double classifyInstance(InstanceContainer instance) throws Exception
    {
        if (!this.m_classifierExist) {
            if (this.m_firstClassSeen != NO_CLASS) {
                return this.m_firstClassSeen;
            } else {
                return Instance.missingValue();
            }
        }
        if (instance instanceof Instance) {
            Instance inst;
            inst = (Instance)instance;
            return this.m_svm.classifyInstance(inst);
        }
        return this.m_svm.classifyInstance(instance);
    }

    public void startActiveLearningSession() throws Exception
    {
        this.m_classifierExist = false;
        this.m_firstClassSeen = NO_CLASS;

        this.m_data = null;
        this.m_wasQueried = null;
        this.m_trainingSet = null;
        this.m_fullTrainingSet = null;
        this.m_pool = null;
        this.m_holdOut = new FastVector();
        // random
        if (this.m_rndm == null) {
            this.m_rndm = new Random(m_seed);
        }
        // progress
        if (m_selfTestInfo == null) {
            m_selfTestInfo = new FastVector();
        }
        if (this.m_mode == MODE_CEM_PROOF) {
            if (m_selfTestPosRatio == null) {
                m_selfTestPosRatio = new FastVector();
            }
            if (m_selfTestActualPosRatio == null) {
                m_selfTestActualPosRatio = new FastVector();
            }
            if (m_selfTestSVM == null) {
                m_selfTestSVM = new FastVector();
            }
        }
        this.m_lastEntropyRes = 0;
        this.m_previousEntropyRes = 0;
        this.m_lastFullEntropyRes = 0;
        this.m_previousFullEntropyRes = 0;
    }

    private double paramSelection(DataContainer train) throws Exception
    {
        int i;
        double[] parArr = m_gammaMS;
        double suc, bestSuc=-Double.MAX_VALUE;
        int bestParamIndx=-1;

        if (this.m_mode == MODE_ACTIVE_CEM) {
            System.out.println("!!!!!!! choosing with entropy");
        } else {
            if (this.m_mode != MODE_ACTIVE_LOO) throw new Exception("Unknwon mode value");
            if (train.size() <= 20) {
                System.out.println("!!!!!!! choosing with leave one out ");
            } else {
                System.out.println("!!!!!!! choosing with 10-fold CV ");
            }
        }
        for (i=0; i<parArr.length; i++) {
            if (this.m_mode == MODE_ACTIVE_CEM) {
                suc = this.entropyCalc(train, parArr[i]);
            } else {
                if (train.size() <= 20) {
                    suc = this.leaveOneOut(train, parArr[i]);
                } else {
                    suc = this.KFoldCV(train,
                                       10,
                                       parArr[i],
                                       this.m_rndm);
                }
            }
            System.out.println("param=" + parArr[i] + " suc=" + (suc*100));
            if (Utils.gr(suc, bestSuc)) {
                bestSuc = suc;
                bestParamIndx = i;
            }
        }
        this.setGamma(parArr[bestParamIndx]);
        System.out.println("BestParam=" + parArr[bestParamIndx]);
        return parArr[bestParamIndx];
    }

    private double sucOnTestSet(DataContainer train, double par) throws Exception
    {
        int i;
        InstanceContainer ins;
        int suc, size = this.m_testSet.size();
        int cls, correctCls;

        this.setParam(par);
        this.m_svm.buildClassifier(train);
        suc = 0;
        for (i=0; i<size; i++) {
            ins = this.m_testSet.getInstance(i);
            cls = (int)(this.m_svm.classifyInstance(ins));
            correctCls = (int)(ins.classValue());
            if (cls == correctCls) {
                suc++;
            }
        }
        return (((double)suc)/((double)size));
    }

    protected boolean isSucOnTrainSet(DataContainer train) throws Exception
    {
        int i;
        InstanceContainer ins;
        int size = train.size();
        int cls, correctCls;
        boolean end = true;

        for (i=0; i<size; i++) {
            ins = train.getInstance(i);
            cls = (int)(this.m_svm.classifyInstance(ins));
            correctCls = (int)(ins.classValue());
            if (cls != correctCls) {
                end = false;
                break;
            }
        }
        return end;
    }

    private double sucOnOrigTestSet(Classifier classifier) throws Exception
    {
        int i;
        Instance ins;
        int suc, size = this.m_origTestSet.size();
        int cls, correctCls;

        suc = 0;
        for (i=0; i<size; i++) {
            ins = this.m_origTestSet.instance(i);
            cls = (int)(classifier.classifyInstance(ins));
            correctCls = (int)(ins.classValue());
            if (cls == correctCls) {
                suc++;
            }
        }
        return (((double)suc)/((double)size));
    }

    private DataContainer sampleForGamma(DataContainer data, Instances origData, double ratio, Random rnd)
    {
        int i, indx, size = data.size();
        int sampleSize = (int)(ratio * ((double)data.size()));
        DataContainer sample = data.makeEmptyCopy();
        InstanceContainer ins;

        for (i=0; i<sampleSize; i++) {
            indx = (int)(Math.ceil(data.size() * rnd.nextDouble()))-1;
            ins = data.getInstance(indx);
            sample.addInstance(ins);
            data.deleteInstance(indx);
            if (origData != null) {
                origData.deleteInstance(indx);
            }
        }
        return sample;
    }

    private void addPoolToData(DataContainer data)
    {
        int i, poolSz = this.m_pool.size();
        InstanceContainer ins;
        for (i=0; i<poolSz; i++) {
            ins = this.m_pool.getInstance(i);
            data.addInstance(ins);
        }
    }

    public void setDataAndPool(Instances data, DataContainer pool) throws Exception
    {
        this.initIndexesAndFlags(pool);
    }

    private void initIndexesAndFlags(DataContainer pool)
    {
        // initialize the indexes
        InstanceContainer ins;
        int size = pool.size();
        // the pool is the intire data set in this session
        // indexize all the data and init the wasQueried flags
        this.m_wasQueried = new int[size];
        for(int i=0; i<size; i++) {
            ins = this.m_data.getInstance(i);
            ins.setIndex(i);
            ins = pool.getInstance(i);
            ins.setIndex(i);
            this.m_wasQueried[i] = UNCLASSIFIED;
        }
    }

    protected int instanceToQueryInitMode(DataContainer pool) throws Exception
    {
        FastVector indexes = new FastVector();
        int res = -1;

        if (this.m_firstClassSeen == NO_CLASS) {
            return 0;
        } else if (!this.m_classifierExist) {
            int i;
            int size = pool.size();
            InstanceContainer ins;
            for (i=0; i<size; i++) {
                ins = pool.getInstance(i);
                if (((int)ins.classValue()) != this.m_firstClassSeen) {
                    res = i;
                    break;
                }
            }
            if (res==-1) throw new Exception("only one class found");
            return res;
        }
        return -1;
    }

    private void updateNewQueryResult(DataContainer data) throws Exception
    {
        InstanceContainer newIns = data.getInstance(data.size()-1);
        int newIndx = newIns.getIndex();
        int newClass = (int)(newIns.classValue());
        this.m_wasQueried[newIndx] = newClass;
    }

    protected boolean betterInInfoGain(DataContainer data, DataContainer pool) throws Exception
    {
        boolean res = true;
        int i, poolSize = pool.size(), trnSize = data.size();
        InstanceContainer ins;
        int[] partition = new int[this.m_data.size()];

        if (trnSize+poolSize != this.m_data.size()) throw new Exception("trnSize=" + trnSize + " poolSize=" + poolSize + " total=" + this.m_data.size());
        for (i=0; i<partition.length; i++) {
            partition[i] = UNCLASSIFIED;
        }
        for (i=0; i<poolSize; i++) {
            ins = pool.getInstance(i);
            partition[ins.getIndex()] = (int)(this.classifyInstance(ins));
        }
        int numPos=0, numNeg=0;
        double ratio, entropy;
        double ln2 = Math.log(2);
        for (i=0; i<partition.length; i++) {
            if (partition[i] == UNCLASSIFIED) continue;
            if (partition[i] == 0) numNeg++;
            else numPos++;
        }
        ratio = ((double)numPos)/((double)(numNeg + numPos));
        System.out.println("posRatio: " + ratio);
        entropy = 0.0;
        if (!Utils.eq(ratio, 0.0)) {
            entropy = -1.0 * ratio * (Math.log(ratio)/ln2);
        }
        ratio = 1.0 - ratio;
        if (!Utils.eq(ratio, 0.0)) {
            entropy += (-1.0 * ratio * (Math.log(ratio)/ln2));
        }
        System.out.println("entropy=" + entropy);
        this.m_previousEntropyRes = this.m_lastEntropyRes;
        this.m_lastEntropyRes = entropy;
        ((FastVector)(m_selfTestInfo.elementAt(data.size()-1))).addElement(new Double(entropy*100));

        for (i=0; i<trnSize; i++) {
            ins = data.getInstance(i);
            partition[ins.getIndex()] = (int)(this.classifyInstance(ins));
        }
        for (i=0; i<partition.length; i++) {
            if (partition[i] == UNCLASSIFIED) throw new Exception("There is an unclassified instance");
        }
        numPos=numNeg=0;
        for (i=0; i<partition.length; i++) {
            if (partition[i] == 0) numNeg++;
            else numPos++;
        }
        ratio = ((double)numPos)/((double)(numNeg + numPos));
        entropy = 0.0;
        if (!Utils.eq(ratio, 0.0)) {
            entropy = -1.0 * ratio * (Math.log(ratio)/ln2);
        }
        ratio = 1.0 - ratio;
        if (!Utils.eq(ratio, 0.0)) {
            entropy += (-1.0 * ratio * (Math.log(ratio)/ln2));
        }
        this.m_previousFullEntropyRes = this.m_lastFullEntropyRes;
        this.m_lastFullEntropyRes = entropy;

        if (this.m_mode == MODE_CEM_PROOF) {
            System.out.println("$$$$$$$$$$$$$$ MODE_CEM_PROOF $$$$$$$$$$$$$$$");
            System.out.println("TrainSz=" + trnSize + " poolSz=" + poolSize);
            ratio = ((double)numNeg)/((double)(numNeg + numPos));
            ((FastVector)(m_selfTestSVM.elementAt(data.size()-1))).addElement(new Double(ratio*100));
            System.out.println("1. SVM pool ratio: " + ratio);

            numPos=numNeg=0;
            for (i=0; i<trnSize; i++) {
                ins = data.getInstance(i);
                if (((int)ins.classValue())==0) numNeg++;
                else numPos++;
            }
            ratio = ((double)numNeg)/((double)(numNeg + numPos));
            ((FastVector)(m_selfTestPosRatio.elementAt(data.size()-1))).addElement(new Double(ratio*100));
            System.out.println("2. Chosen training set ratio: " + ratio);
            for (i=0; i<poolSize; i++) {
                ins = pool.getInstance(i);
                if (((int)ins.classValue())==0) numNeg++;
                else numPos++;
            }
            ratio = ((double)numNeg)/((double)(numNeg + numPos));
            ((FastVector)(m_selfTestActualPosRatio.elementAt(data.size()-1))).addElement(new Double(ratio*100));
            System.out.println("3. Total training and pool ratio: " + ratio);
            System.out.println("$$$$$$$$$$$$$$ MODE_CEM_PROOF END $$$$$$$$$$$$$$$");
        }
        return res;
    }

    protected boolean betterInSelfEstimate(DataContainer data, DataContainer pool) throws Exception
    {
        double suc;
        boolean res = true;
        if (data.size() <= 20) {
            suc = this.leaveOneOut(data, this.m_svm.getGamma());
        } else {
            suc = this.KFoldCV(data, 10, this.m_svm.getGamma(), this.m_rndm);
        }
        ((FastVector)(m_selfTestInfo.elementAt(data.size()-1))).addElement(new Double(suc*100));
        System.out.println("suc= " + suc);
        this.m_previousEntropyRes = this.m_previousFullEntropyRes = this.m_lastEntropyRes;
        this.m_lastEntropyRes = this.m_lastFullEntropyRes = suc;
        return res;
    }

    protected int translateFromDataToPoolIndex(int dataIndex, DataContainer pool) throws Exception
    {
        int i, id;
        int size = pool.size();
        for (i=0; i<size; i++) {
            id = pool.getInstance(i).getIndex();
            if (id == dataIndex) {
                return i;
            }
        }
        throw new Exception("Error in SvmClusterSimple: in translateFromDataToPoolIndex index not found.");
    }

    protected InstanceContainer translateFromDataToPoolInstance(int dataIndex, DataContainer pool) throws Exception
    {
        int i, id;
        InstanceContainer ins;
        int size = pool.size();
        for (i=0; i<size; i++) {
            ins = pool.getInstance(i);
            id = ins.getIndex();
            if (id == dataIndex) {
                return ins;
            }
        }
        throw new Exception("Error in SvmClusterSimple: in translateFromDataToPoolInstance index not found.");
    }

    protected DataContainer sampleFromPool(DataContainer pool,
                                           double ratio,
                                           Random rnd)
    {
        int sampleSize = (int)(ratio * ((double)pool.size()));
        return this.sampleFromPool(pool,
                                   sampleSize,
                                   rnd,
                                   null);
    }

    protected DataContainer sampleFromPool(DataContainer pool,
                                           int sampleSize,
                                           Random rnd,
                                           int[] indexes)
    {
        int i, indx1, indx2, poolSz = pool.size();
        DataContainer sample = pool.makeEmptyCopy();
        InstanceContainer ins;
        FastVector chooseFrom = new FastVector();

        for (i=0; i<poolSz; i++) {
            chooseFrom.addElement(new Integer(i));
        }
        for (i=0; i<sampleSize; i++) {
            indx1 = (int)(Math.ceil(chooseFrom.size() * rnd.nextDouble()))-1;
            indx2 = ((Integer)(chooseFrom.elementAt(indx1))).intValue();
            chooseFrom.removeElementAt(indx1);
            ins = pool.getInstance(indx2);
            sample.addInstance(ins);
            if (indexes != null) {
                indexes[i] = indx2;
            }
        }
        return sample;
    }

    private void modelSelection(DataContainer data) throws Exception
    {
        // MODE_MS_BEST=3, MODE_MS_LOO=4, MODE_MS_CEM=5,  MODE_MS_WORST=6,
        // MODE_MS_BEST_ALL=7, MODE_MS_WORST_ALL=10, MODE_MS_CEM_ALL=9, MODE_MS_LOO_ALL=8
        int i, stopNum = 20;
        double[] parArr = m_gammaMS;
        if (data.size() == stopNum) {
            int size;
            String fileSuf;
            boolean toAdd = false;

            switch (this.m_mode) {
                case MODE_MS_BEST:
                    fileSuf = new String("best_" + stopNum);
                    break;
                case MODE_MS_LOO:
                    fileSuf = new String("loo_" + stopNum);
                    break;
                case MODE_MS_CEM:
                    fileSuf = new String("cem_" + stopNum);
                    break;
                case MODE_MS_WORST:
                    fileSuf = new String("worst_" + stopNum);
                    break;
                case MODE_MS_BEST_ALL:
                    fileSuf = new String("bestAll_" + stopNum);
                    break;
                case MODE_MS_LOO_ALL:
                    fileSuf = new String("looAll_" + stopNum);
                    break;
                case MODE_MS_CEM_ALL:
                    fileSuf = new String("cemAll_" + stopNum);
                    break;
                case MODE_MS_WORST_ALL:
                    fileSuf = new String("worstAll_" + stopNum);
                    break;
                default:
                    throw new Exception("unknown mode");
            }
            if (this.m_params == null) {
                String fileName = new String(m_paramsPath + stopNum + "/" + this.m_data.relationName() + fileSuf);
                File f = new File(fileName);
                if (f.exists()) {
                    FileInputStream inFile = new FileInputStream(f);
                    ObjectInputStream inObj = new ObjectInputStream(inFile);
                    m_params = (FastVector)(inObj.readObject());
                    inObj.close();
                    this.m_reading = true;
                    this.m_paramsIndx = 0;
                } else {
                    m_params = new FastVector();
                    this.m_reading = false;
                }
            }
            if (this.m_reading) {
                this.setParam(((Double)(this.m_params.elementAt(this.m_paramsIndx))).doubleValue());
                this.m_paramsIndx++;
            } else {
                // adjust parameters
                double suc, bestSuc=-Double.MAX_VALUE, worstSuc=Double.MAX_VALUE;
                int bestParamIndx=-1;
                for (i=0; i<parArr.length; i++) {
                    switch (this.m_mode) {
                        case MODE_MS_BEST:
                            suc = this.sucOnTestSet(data, parArr[i]);
                            break;
                        case MODE_MS_LOO:
                            suc = this.leaveOneOut(data, parArr[i]);
                            break;
                        case MODE_MS_CEM:
                            suc = this.entropyCalc(data, parArr[i]);
                            break;
                        case MODE_MS_WORST:
                            suc = this.sucOnTestSet(data, parArr[i]);
                            break;
                        case MODE_MS_BEST_ALL:
                            this.addPoolToData(data);
                            suc = this.sucOnTestSet(data, parArr[i]);
                            break;
                        case MODE_MS_LOO_ALL:
                            toAdd = true;
                            suc = this.leaveOneOut(data, parArr[i]);
                            break;
                        case MODE_MS_CEM_ALL:
                            toAdd = true;
                            suc = this.entropyCalc(data, parArr[i]);
                            break;
                        case MODE_MS_WORST_ALL:
                            this.addPoolToData(data);
                            suc = this.sucOnTestSet(data, parArr[i]);
                            break;
                        default:
                            throw new Exception("unknown mode");
                    }
                    System.out.println("param=" + parArr[i] + " suc=" + (suc*100));
                    if (this.m_mode == MODE_MS_WORST) {
                        if (Utils.sm(suc, worstSuc)) {
                            worstSuc = suc;
                            bestParamIndx = i;
                        }
                    } else {
                        if (Utils.gr(suc, bestSuc)) {
                            bestSuc = suc;
                            bestParamIndx = i;
                        }
                    }
                }
                this.setParam(parArr[bestParamIndx]);
                System.out.println("BestParam=" + parArr[bestParamIndx]);
                this.m_params.addElement(new Double(parArr[bestParamIndx]));
                String fileName = new String(m_paramsPath + stopNum + "/" + this.m_data.relationName() + fileSuf);
                File f = new File(fileName);
                FileOutputStream outFile = new FileOutputStream(f);
                ObjectOutputStream outObj = new ObjectOutputStream(outFile);
                outObj.writeObject(this.m_params);
                outObj.close();
            }
            if (toAdd) {
                this.addPoolToData(data);
            }
        }
    }

    private double entropyCalc(DataContainer train, double par) throws Exception
    {
        int i, poolSize = this.m_pool.size(), trnSize = train.size();
        InstanceContainer ins;
        int[] partition = new int[poolSize + trnSize];

        for (i=0; i<partition.length; i++) {
            partition[i] = UNCLASSIFIED;
        }
        this.setParam(par);
        this.m_svm.buildClassifier(train);
        for (i=0; i<poolSize; i++) {
            ins = this.m_pool.getInstance(i);
            partition[ins.getIndex()] = (int)(this.classifyInstance(ins));
        }
        for (i=0; i<trnSize; i++) {
            ins = train.getInstance(i);
            partition[ins.getIndex()] = (int)(this.classifyInstance(ins));
        }
        for (i=0; i<partition.length; i++) {
            if (partition[i] == UNCLASSIFIED) throw new Exception("There is an unclassified instance");
        }

        int numPos=0, numNeg=0;
        double ratio, entropy;
        double ln2 = Math.log(2);
        for (i=0; i<partition.length; i++) {
            if (partition[i] == UNCLASSIFIED) continue;
            if (partition[i] == 0) numNeg++;
            else numPos++;
        }
        ratio = ((double)numPos)/((double)(numNeg + numPos));
        entropy = 0.0;
        if (!Utils.eq(ratio, 0.0)) {
            entropy = -1.0 * ratio * (Math.log(ratio)/ln2);
        }
        ratio = 1.0 - ratio;
        if (!Utils.eq(ratio, 0.0)) {
            entropy += (-1.0 * ratio * (Math.log(ratio)/ln2));
        }
        return entropy;
    }

    public boolean MulticlassCanISwitch()
    {
        return true;
    }

    protected void setGamma(double gamma)
    {
        this.m_svm.setKernelType(svm_parameter.RBF);
        this.m_svm.setGamma(gamma);
    }

    protected void setDegree(double degree)
    {
        this.m_svm.setKernelType(svm_parameter.POLY);
        this.m_svm.setDegree(degree);
    }

    protected void setParam(double p)
    {
        if (!this.m_polynomial) {
            this.setGamma(p);
        } else {
            this.setDegree(p);
        }
    }

    private void setPolynomial(boolean f)
    {
        m_polynomial = f;
    }

    protected void setSvmParam(SvmLib svm, double p)
    {
        if (!this.m_polynomial) {
            svm.setKernelType(svm_parameter.RBF);
            svm.setGamma(p);
        } else {
            svm.setKernelType(svm_parameter.POLY);
            svm.setDegree(p);
        }
    }

    public void setSeed(int seed)
    {
        this.m_seed = seed;
    }

    protected void waitForUser()
    {
        if (this.m_waitForUserFlag) {
            try {
                System.out.println("Press Enter to continue.");
                System.in.read();
                System.in.read();
            } catch (Exception e) {
            }
        }
    }

    public void setWaitForUser(boolean f)
    {
        this.m_waitForUserFlag = f;
    }

    protected void printMeg(String msg)
    {
        System.out.println(msg);
        this.waitForUser();
    }

    protected double KFoldCV(DataContainer data,
                             int numFolds,
                             double p,
                             Random rnd) throws Exception
    {
        int i, j, k;
        int indx, size=data.size();
        int suc, currCls, correctCls;
        SvmLib svm;
        DataContainer[] folds = new DataContainer[numFolds];
        DataContainer trainData, trainCopy;
        InstanceContainer ins;

        // divide
        for (i=0; i<numFolds; i++) {
            folds[i] = data.makeEmptyCopy();
        }
        trainCopy = data.makeEmptyCopy();
        for (i=0; i<size; i++) {
            ins = data.getInstance(i);
            trainCopy.addInstance(ins);
        }
        j=0;
        for (i=0; i<size; i++) {
            indx = (int)(Math.ceil(trainCopy.size() * rnd.nextDouble()))-1;
            ins = trainCopy.getInstance(indx);
            folds[j].addInstance(ins);
            trainCopy.deleteInstance(indx);
            j++;
            if (j == numFolds) j=0;
        }
        // learn & test
        suc = 0;
        for (j=0; j<numFolds; j++) {
            trainData = data.makeEmptyCopy();
            for (i=0; i<numFolds; i++) {
                if (i != j) {
                    size = folds[i].size();
                    for (k=0; k<size; k++) {
                        ins = folds[i].getInstance(k);
                        trainData.addInstance(ins);
                    }
                }
            }
            svm = new SvmLib();
            this.setSvmParam(svm, p);
            svm.buildClassifier(trainData);
            // test
            size = folds[j].size();
            for (i=0; i<size; i++) {
                ins = folds[j].getInstance(i);
                currCls = (int)(svm.classifyInstance(ins));
                correctCls = (int)(ins.classValue());
                if (currCls == correctCls) {
                    suc++;
                }
            }
        }
        size = data.size();
        return ((double)suc)/((double)size);
    }

    protected double holdOut(DataContainer data,
                             double p,
                             Random rnd) throws Exception
    {
        int i, indx, size=data.size();
        int suc, currCls, correctCls;
        SvmLib svm;
        DataContainer trainData, testData;
        InstanceContainer ins;

        // divide
        testData = data.makeEmptyCopy();
        for (i=0; i<size; i++) {
            ins = data.getInstance(i);
            testData.addInstance(ins);
        }
        size = size/2;
        trainData = data.makeEmptyCopy();
        for (i=0; i<size; i++) {
           indx = (int)(Math.ceil(testData.size() * rnd.nextDouble()))-1;
           ins = testData.getInstance(indx);
           trainData.addInstance(ins);
           testData.deleteInstance(indx);
        }
        // learn
        svm = new SvmLib();
        this.setSvmParam(svm, p);
        svm.buildClassifier(trainData);
        // test
        suc = 0;
        size = testData.size();
        for (i=0; i<size; i++) {
            ins = testData.getInstance(i);
            currCls = (int)(svm.classifyInstance(ins));
            correctCls = (int)(ins.classValue());
            if (currCls == correctCls) {
                suc++;
            }
        }
        return ((double)suc)/((double)size);
    }

    protected double leaveOneOut(DataContainer data, double p) throws Exception
    {
        int i, j, size=data.size();
        int suc, currCls, correctCls;
        SvmLib svm;
        DataContainer trainData;
        InstanceContainer ins;

        suc=0;
        for (i=0; i<size; i++) {
            trainData = data.makeEmptyCopy();
            for (j=0; j<size; j++) {
                if (i != j) {
                    ins = data.getInstance(j);
                    trainData.addInstance(ins);
                }
            }
            svm = new SvmLib();
            this.setSvmParam(svm, p);
            svm.buildClassifier(trainData);
            ins = data.getInstance(i);
            currCls = (int)(svm.classifyInstance(ins));
            correctCls = (int)(ins.classValue());
            if (currCls == correctCls) {
                suc++;
            }
        }
        return ((double)suc)/((double)size);
    }

    public void setMode(int mode)
    {
        this.m_mode = mode;
    }

    public void setSampleRatio(double r)
    {
        this.m_sampleRatio = r;
    }

    protected void initIndexes(DataContainer pool)
    {
        // initialize the indexes
        InstanceContainer ins;
        int size = pool.size();
        for(int i=0; i<size; i++) {
            ins = pool.getInstance(i);
            ins.setIndex(i);
        }
    }

    protected void unsetIndexes(DataContainer pool)
    {
        InstanceContainer ins;
        int size = pool.size();
        for(int i=0; i<size; i++) {
            ins = pool.getInstance(i);
            ins.setIndex(-1);
        }
    }

    protected FastVector saveIndexes(DataContainer pool)
    {
        FastVector indexes = new FastVector();
        InstanceContainer ins;
        int size = pool.size();
        for(int i=0; i<size; i++) {
            ins = pool.getInstance(i);
            indexes.addElement(new Integer(ins.getIndex()));
        }
        return indexes;
    }

    protected void loadIndexes(DataContainer pool, FastVector indexes)
        throws Exception
    {
        int indx;
        InstanceContainer ins;
        int size = pool.size();
        if (indexes.size() != size) throw new Exception("loadIndexes - sizes do not match");
        for(int i=0; i<size; i++) {
            ins = pool.getInstance(i);
            indx = ((Integer)(indexes.elementAt(i))).intValue();
            ins.setIndex(indx);
        }
    }

    public void setOptions(String options[]) throws Exception
    {
        String IString = Utils.getOption('I', options);
        if (IString.length() != 0) {
            this.setWaitForUser(true);
        }
        String MString = Utils.getOption('M', options);
        if (MString.length() != 0) {
            this.setMode((new Integer(MString)).intValue());
        }
        String RString = Utils.getOption('R', options);
        if (RString.length() != 0) {
            this.setSampleRatio((new Double(RString)).doubleValue());
        }
        String PString = Utils.getOption('P', options);
        if (PString.length() != 0) {
            this.setPolynomial(true);
        }
    }

    public void setSvmOptions(String[] options) throws Exception
    {
        this.m_svm.setOptions(options);
    }
}
