package weka.active;

import java.util.*;
import java.io.*;
import weka.active.*;
import weka.clusterers.FarthestKernelSampler;
import weka.classifiers.libsvm.*;
import weka.core.*;

public class SvmMAB3Info extends weka.active.SvmActiveLearner
                         implements RandomUser, OptionHandler
{
    private static final int SIMPLE = 0;
    private static final int KFAR = 1;
    private static final int ERROR_REDUCTION = 2;  
    
    private int m_minSampleSize = 10;
    private int m_maxSampleSize = 100;
    private int m_currSampleSize;
    private int m_dec = 1;
    
    private MultiArmBandit2 m_mab = null;
    private boolean m_mabRndmFlag = false;
    private double[] m_chosenProbs;
    // params 
    private int m_initialRandom = 0;    
    private double m_distLimit = 0.05; 
    private double m_scaling = 100;
    
    public SvmMAB3Info() 
    {}
    
    public void startActiveLearningSession() throws Exception
    {  
        int i;
        
        super.startActiveLearningSession();                        
        this.m_mab = null;
        this.m_chosenProbs = new double[3];
        for (i=0; i<this.m_chosenProbs.length; i++) {
            this.m_chosenProbs[i] = 0.0;
        }
        this.m_currSampleSize = m_maxSampleSize;
    }                
    
    public int instanceToQuery(DataContainer pool) throws Exception
    {
        int res = super.instanceToQuery(pool);             
        if (res != -1) {
            return res;
        }      
        if (this.m_fullTrainingSet.size() < this.m_initialRandom) {
            res = this.instanceToQueryRandomMode(pool);
            if (res == -1) {
                throw new Exception("Random querying failed");                   
            }
            return res;
        }
        res = this.instanceToQueryMAB3(pool);        
        if (res != -1) {                              
            return res;                               
        }        
        throw new Exception("no res");
    }
    
    public double[] distributionToQuerySimple(DataContainer pool) throws Exception
    {
        int i;
        int size = pool.size();
        InstanceContainer ins;        
        double currConf, sum;  
        double lowestConf = Double.MAX_VALUE, highestConf=-Double.MAX_VALUE;
        double[] prob = new double[size];
                
        System.out.println("simple dist");
        // find all instances confidence and scale by min-max      
        for (i=0; i<size; i++) {
            ins = pool.getInstance(i);
            currConf = Math.abs(this.m_svm.confindenceOnInstance(ins));
            prob[i] = currConf;              
            if (Utils.sm(prob[i], lowestConf)) {
                lowestConf = prob[i];
            }
            if (Utils.gr(prob[i], highestConf)) {
                highestConf = prob[i];
            }                  
        }                
        for (i=0; i<size; i++) {
            if (Utils.eq(lowestConf, highestConf)) {
                prob[i] = 0.5;
            } else {                            
                prob[i] = (prob[i] - lowestConf) / (highestConf - lowestConf);
                if (Utils.sm(prob[i], 0.0)) {
                    prob[i] = 0.0;
                } 
                if (Utils.gr(prob[i], 1.0)) {
                    prob[i] = 1.0;
                }
            }            
        }              
        // scale up             
        double scale = this.m_scaling; // PARAMETER         
        for (i=0; i<size; i++) {
            prob[i] *= scale;            
        }              
        // exponential scale 
        sum = 0.0;        
        for (i=0; i<size; i++) {
            prob[i] = Math.exp(-1.0 * prob[i]);            
            sum += prob[i];
        }                
        for (i=0; i<size; i++) {
            prob[i] = prob[i] / sum;              
        }
        for (i=0; i<size; i++) {
            if (Utils.grOrEq(prob[i], 0.01)) {                
                System.out.println("prob[" + i + "]=" + prob[i]);
            }            
        }        
        // check
        sum = 0.0;
        for (i=0; i<size; i++) {            
            sum += prob[i];
        } 
        if (!Utils.eq(sum, 1.0)) throw new Exception("sum probs is not 1.0 - sum=" + sum);
        return prob;
    }
    
    public double[] distributionToQueryKfar(DataContainer pool) throws Exception
    {
        int i; 
        int trnSize = this.m_fullTrainingSet.size();                
        InstanceContainer pIns;
        
        // create sampler
        FarthestKernelSampler sampler = new FarthestKernelSampler();          
        // we will take to init all the points that were allready queried
        FastVector initSampler = new FastVector();        
        for (i=0; i<trnSize; i++) {
            pIns = this.m_fullTrainingSet.getInstance(i);
            initSampler.addElement(pIns);
        }
        return sampler.nextDistribution(this.m_svm.getParams(), 
                                        pool, 
                                        initSampler,
                                        this.m_scaling);        
    }
    
    public double[] distributionToQueryErrorReduction(DataContainer pool) throws Exception    
    {        
        int i, newIndx, currCls;
        int sampleSz, poolSz = pool.size();
        InstanceContainer ins;
        DataContainer sampledPool;
        double currConf, currLoss;
        double sum, lowestLoss = Double.MAX_VALUE, highestLoss=-Double.MAX_VALUE;
        double[] prob = new double[poolSz];        
        int[] indexes;
        
        System.out.println("error-reduction dist");
        for (i=0; i<poolSz; i++) {
            prob[i] = 0.0;            
        }        
        if (m_currSampleSize <= m_minSampleSize) {
            sampleSz = m_minSampleSize; 
            m_currSampleSize = m_minSampleSize;
        } else {
            sampleSz = m_currSampleSize;
            m_currSampleSize -= m_dec;
        } 
        if (poolSz < sampleSz) {
            sampleSz = poolSz; 
            sampledPool = pool;
            indexes = new int[sampleSz];
            for (i=0; i<sampleSz; i++) {
                indexes[i] = i;            
            }
        } else {            
            indexes = new int[sampleSz];
            sampledPool = this.sampleFromPool(pool, 
                                              sampleSz, 
                                              this.m_rndm,
                                              indexes);            
        }
        System.out.println("Sample size=" + sampleSz);
        
        // find the instance with lowest confidence                
        for (i=0; i<sampleSz; i++) {
            currLoss = 0.0;
            ins = sampledPool.getInstance(i);
            currConf = Math.abs(this.m_svm.confindenceOnInstance(ins));            
            currConf = this.logisticRegression(currConf);            
            currCls = (int)(this.m_svm.classifyInstance(ins));
            currLoss += (this.expectedLogLoss(ins, currCls, sampledPool) * currConf);            
            currCls = ((currCls == 0) ? 1 : 0);
            currConf = 1.0 - currConf;
            currLoss += (this.expectedLogLoss(ins, currCls, sampledPool) * currConf);   
            
            newIndx = indexes[i];
            prob[newIndx] = currLoss;            
            if (Utils.sm(prob[newIndx], lowestLoss)) {
                lowestLoss = prob[newIndx];
            }
            if (Utils.gr(prob[newIndx], highestLoss)) {
                highestLoss = prob[newIndx];
            }                  
        }                
        for (i=0; i<sampleSz; i++) {
            newIndx = indexes[i];
            if (Utils.eq(lowestLoss, highestLoss)) {                
                prob[newIndx] = 0.5;
            } else {                            
                prob[newIndx] = (prob[newIndx] - lowestLoss) / (highestLoss - lowestLoss);
                if (Utils.sm(prob[newIndx], 0.0)) {
                    prob[newIndx] = 0.0;
                } 
                if (Utils.gr(prob[newIndx], 1.0)) {
                    prob[newIndx] = 1.0;
                }
            }            
        }              
        // scale up             
        double scale = this.m_scaling; // PARAMETER        
        for (i=0; i<sampleSz; i++) {
            newIndx = indexes[i];
            prob[newIndx] *= scale;           
        }              
        // exponential scale 
        sum = 0.0;        
        for (i=0; i<sampleSz; i++) {
            newIndx = indexes[i];
            prob[newIndx] = Math.exp(-1.0 * prob[newIndx]);
            sum += prob[newIndx];                        
        }              
        for (i=0; i<sampleSz; i++) {
            newIndx = indexes[i];
            prob[newIndx] = prob[newIndx] / sum;              
        }        
        for (i=0; i<sampleSz; i++) {
            newIndx = indexes[i];
            if (Utils.grOrEq(prob[newIndx], 0.01)) {                
                System.out.println("prob[" + newIndx + "]=" + prob[newIndx]);
            }            
        }        
        // check
        sum = 0.0;
        for (i=0; i<poolSz; i++) {             
            sum += prob[i];
        } 
        if (!Utils.eq(sum, 1.0)) throw new Exception("sum probs is not 1.0 - sum=" + sum);
        return prob;                       
    }
    
    private double expectedLogLoss(InstanceContainer insToAdd, int cls, DataContainer pool) throws Exception
    {
        int i, poolSz = pool.size();
        double currConf, sum;
        InstanceContainer ins;
        
        insToAdd.setClassValue(cls);
        this.m_fullTrainingSet.addInstance(insToAdd);
        this.m_svm.buildClassifier(this.m_fullTrainingSet);
        sum = 0.0;
        for (i=0; i<poolSz; i++) {
            ins = pool.getInstance(i);
            currConf = Math.abs(this.m_svm.confindenceOnInstance(ins));
            currConf = this.logisticRegression(currConf);
            if (!Utils.eq(currConf, 0)) {                
                sum += (currConf * Math.log(currConf));
            }
            currConf = 1.0 - currConf;
            if (!Utils.eq(currConf, 0)) {                
                sum += (currConf * Math.log(currConf));
            }
            if (Utils.gr(sum, 0.0)) throw new Exception("sum is positive");
        }
        sum /= ((double)poolSz);
        sum = Math.abs(sum);
        this.m_fullTrainingSet.deleteInstance(this.m_fullTrainingSet.size()-1);
        this.m_svm.buildClassifier(this.m_fullTrainingSet);
        return sum;
    }
    
    private double logisticRegression(double conf) throws Exception
    {
        if (Utils.sm(conf, 0.0)) throw new Exception("conf is negative" + conf);
        double res = (1.0 / (1.0 + Math.exp(-1.0 * conf)));
        if (Utils.sm(res, 0.5) || Utils.gr(res, 1.0)) throw new Exception("logic reg is not in [0.5, 1]");
        return res;
    }
    
    private double getLastReward() throws Exception
    {                
        double reward;
        // calc reward - can be in the range of [1-e, e-1] (as entropy is in [0,1])
        reward = Math.exp(this.m_lastEntropyRes) - Math.exp(this.m_previousEntropyRes);
        // normalize 
        reward = (reward - 1.0 + Math.E) / ((2.0 * Math.E) - 2.0);
        if (Utils.sm(reward, 0.0) || Utils.gr(reward, 1.0)) {
            throw new Exception("reward out of range=" + reward + " last=" + this.m_lastEntropyRes + " prev=" + this.m_previousEntropyRes);
        }
        return reward;
    }
        
    private int instanceToQueryMAB3(DataContainer pool) throws Exception
    {     
        int i, indx, chosenIndx=-1;
        int numAlgos = 3;
        double reward;
        double[][] dist = new double[numAlgos][];        
        double[][] dist2 = new double[numAlgos][];       
        FastVector indexes;
        double limit = this.m_distLimit; 
                        
        dist[0] = this.distributionToQuerySimple(pool);            
        dist[1] = this.distributionToQueryKfar(pool); 
        dist[2] = this.distributionToQueryErrorReduction(pool);        
        indexes = new FastVector();
        while (indexes.size() == 0) {
            for(i=0; i<dist[0].length; i++) {
                if (Utils.grOrEq(dist[0][i], limit) || 
                    Utils.grOrEq(dist[1][i], limit) ||
                    Utils.grOrEq(dist[2][i], limit)) 
                {
                    indexes.addElement(new Integer(i));
                }
            }  
            limit /= 2;
        }
        dist2[0] = new double[indexes.size()];
        dist2[1] = new double[dist2[0].length];
        dist2[2] = new double[dist2[0].length];
        for (i=0; i<dist2[0].length; i++) {
            indx = ((Integer)(indexes.elementAt(i))).intValue();
            dist2[0][i] = dist[0][indx];
            dist2[1][i] = dist[1][indx];
            dist2[2][i] = dist[2][indx];
        }        
        System.out.println("K=" + dist2[0].length);
        
        if (this.m_mab == null) {           
            this.m_mab = new MultiArmBandit2(m_mabRndmFlag);
            this.m_mab.start(3, this.m_rndm);                         
        } else {
            //if (Utils.eq(this.m_chosenProbs[0], 0.0)) throw new Exception("chosen prob is 0.0");
            reward = this.getLastReward();
            this.m_mab.updateReward(reward, this.m_chosenProbs);            
        }                
        indx = this.m_mab.chooseAction(dist2);        
        chosenIndx = ((Integer)(indexes.elementAt(indx))).intValue();        
        if ((chosenIndx < 0) || (chosenIndx > pool.size()-1)) throw new Exception("chosenIndx out of range. poolsize=" + pool.size() + " indx=" + chosenIndx);
        for (i=0; i<numAlgos; i++) {
            this.m_chosenProbs[i] = dist[i][chosenIndx];
        }
        return chosenIndx;
    }
                               
    private String[] copyOptions(String[] options)
    {
        int i;                
        String[] optionsCopy = new String[options.length];
        for (i=0; i<optionsCopy.length; i++) {
            optionsCopy[i] = new String(options[i]);
        }   
        return optionsCopy;
    }
    
    public void setOptions(String options[]) throws Exception
    {
        int i;
        super.setOptions(options);       
        String SString = Utils.getOption('S', options);
        if (SString.length() != 0) {
            this.setSeed((new Integer(SString)).intValue());            
        }
        String DString = Utils.getOption('D', options);
        if (DString.length() != 0) {
            m_mabRndmFlag = true;            
        }
        String HString = Utils.getOption('H', options);
        if (HString.length() != 0) {
            this.m_initialRandom = (new Integer(HString)).intValue();
        }
        String LString = Utils.getOption('L', options);
        if (LString.length() != 0) {
            this.m_distLimit = (new Double(LString)).doubleValue();
        }
        String AString = Utils.getOption('A', options);
        if (AString.length() != 0) {
            this.m_scaling = (new Double(AString)).doubleValue();            
        }
        String[] svmOptions = Utils.partitionOptions(options);        
        this.m_svm.setOptions(svmOptions);
        Utils.checkForRemainingOptions(options);
    }

    public Enumeration listOptions()
    {
        FastVector newVector = new FastVector(1);        	         
        newVector.addElement(new Option(
         "\tRandom seed.\n"
         +"\t(Default is 1234567)",
         "S", 1, "-S <seed>"));
	return newVector.elements();
    }

    public String[] getOptions()
    {
       String [] options = new String [2];
       int current = 0;
       options[current++] = "-S"; options[current++] = "" + this.m_seed;       
       return options;
    }  
    
    public String toString() 
    {
        StringBuffer msg = new StringBuffer();
        msg.append("SvmSimpleCluster Active Learner\n-----------------------");
        msg.append("The SVM used:\n" + this.m_svm.toString());
        return msg.toString();
    } 
    
    public static void main(String[] args) {
        try {
            System.out.println(ActiveEvaluationPlotter.evaluateModelAndPlot(new SvmMAB3Info(), args));
            ActiveEvaluationPlotter.Plot(m_selfTestInfo, "entropy");                       
        } catch (Exception e) {            
            System.err.println(e.getMessage());            
            e.printStackTrace();
        }  
    }        
}

   