package weka.active;

import java.util.*;
import java.io.*;
import weka.active.*;
import weka.classifiers.libsvm.*;
import weka.core.*;

public class SvmMaxCEMInfo extends weka.active.SvmActiveLearner
                           implements RandomUser, OptionHandler
{            
    private int m_minSampleSize = 10;
    private int m_maxSampleSize = 100;
    private int m_currSampleSize;
    private int m_dec = 1;
    
    public void startActiveLearningSession() throws Exception
    {                  
        super.startActiveLearningSession();                            
        this.m_currSampleSize = m_maxSampleSize;
    } 
    
    public int instanceToQuery(DataContainer pool) throws Exception
    {
        int res = super.instanceToQuery(pool);             
        if (res != -1) {
            return res;
        }
        
        int i, sampleSz, poolSz = pool.size();        
        DataContainer sampledPool;
        int[] indexes;
        
        if (m_currSampleSize <= m_minSampleSize) {
            sampleSz = m_minSampleSize; 
            m_currSampleSize = m_minSampleSize;
        } else {
            sampleSz = m_currSampleSize;
            m_currSampleSize -= m_dec;
        } 
        if (poolSz < sampleSz) {
            sampleSz = poolSz; 
            sampledPool = pool;
            indexes = new int[sampleSz];
            for (i=0; i<sampleSz; i++) {
                indexes[i] = i;            
            }
        } else {            
            indexes = new int[sampleSz];
            sampledPool = this.sampleFromPool(pool, 
                                              sampleSz, 
                                              this.m_rndm,
                                              indexes);            
        }               
        System.out.println("Sample size=" + sampleSz);
                    
        res = this.instanceToQueryMaxCEMMode(sampledPool);
        if (res != -1) {  
            System.out.println("\tquerying in max CEM mode. indx=" + sampledPool.getInstance(res).getIndex());  
            return indexes[res];  
        }        
        throw new Exception("no res");
    }
           
    private int instanceToQueryMaxCEMMode(DataContainer pool) throws Exception
    {        
        int i, best=-1, currCls, poolSz = pool.size();
        InstanceContainer ins;
        double currConf, maxCEM = -Double.MAX_VALUE, currCEM;
        
        if (poolSz == 1) {
            return 0;
        }
        // find the instance with lowest confidence
        for (i=0; i<poolSz; i++) {
            currCEM = 0.0;
            ins = pool.getInstance(i);
            currConf = Math.abs(this.m_svm.confindenceOnInstance(ins));            
            currConf = this.logisticRegression(currConf);            
            currCls = (int)(this.m_svm.classifyInstance(ins));
            currCEM += (this.calcCEM(ins, currCls, pool) * currConf);           
            currCls = ((currCls == 0) ? 1 : 0);
            currConf = 1.0 - currConf;
            currCEM += (this.calcCEM(ins, currCls, pool) * currConf);            
            if (Utils.gr(currCEM, maxCEM)) {
                maxCEM = currCEM;
                best = i;                
            }                           
        }        
        return best;
    }
            
    private double logisticRegression(double conf) throws Exception
    {
        if (Utils.sm(conf, 0.0)) throw new Exception("conf is negative" + conf);
        double res = (1.0 / (1.0 + Math.exp(-1.0 * conf)));
        if (Utils.sm(res, 0.5) || Utils.gr(res, 1.0)) throw new Exception("logic reg is not in [0.5, 1]");
        return res;
    }
    
    private double calcCEM(InstanceContainer insToAdd, int cls, DataContainer pool) throws Exception
    {        
        boolean res = true;        
        int i, indx, indxToIgnore, poolSize = pool.size();                
        InstanceContainer ins;
        int[] partition = new int[this.m_data.size()];
                
        indxToIgnore = insToAdd.getIndex();
        for (i=0; i<partition.length; i++) {
            partition[i] = UNCLASSIFIED;
        }                  
        
        insToAdd.setClassValue(cls);
        this.m_fullTrainingSet.addInstance(insToAdd);
        this.m_svm.buildClassifier(this.m_fullTrainingSet);
        
        for (i=0; i<poolSize; i++) {
            ins = pool.getInstance(i);
            indx = ins.getIndex();
            if (indx == indxToIgnore) continue;
            partition[indx] = (int)(this.classifyInstance(ins));
        }
        int numPos=0, numNeg=0;
        double ratio, entropy;
        double ln2 = Math.log(2);
        for (i=0; i<partition.length; i++) { 
            if (partition[i] == UNCLASSIFIED) continue;
            if (partition[i] == 0) numNeg++;
            else numPos++;
        }        
        ratio = ((double)numPos)/((double)(numNeg + numPos));        
        entropy = 0.0;
        if (!Utils.eq(ratio, 0.0)) {
            entropy = -1.0 * ratio * (Math.log(ratio)/ln2);
        }
        ratio = 1.0 - ratio;
        if (!Utils.eq(ratio, 0.0)) {
            entropy += (-1.0 * ratio * (Math.log(ratio)/ln2));
        }                        
        
        this.m_fullTrainingSet.deleteInstance(this.m_fullTrainingSet.size()-1);
        this.m_svm.buildClassifier(this.m_fullTrainingSet);        
        return entropy;                        
    }
    
    private String[] copyOptions(String[] options)
    {
        int i;                
        String[] optionsCopy = new String[options.length];
        for (i=0; i<optionsCopy.length; i++) {
            optionsCopy[i] = new String(options[i]);
        }   
        return optionsCopy;
    }
    
    public void setOptions(String options[]) throws Exception
    {
        int i;
        super.setOptions(options);       
        String SString = Utils.getOption('S', options);
        if (SString.length() != 0) {
            this.setSeed((new Integer(SString)).intValue());            
        }
        String[] svmOptions = Utils.partitionOptions(options);        
        this.m_svm.setOptions(svmOptions);
        Utils.checkForRemainingOptions(options);
    }

    public Enumeration listOptions()
    {
        FastVector newVector = new FastVector(1);        	         
        newVector.addElement(new Option(
         "\tRandom seed.\n"
         +"\t(Default is 1234567)",
         "S", 1, "-S <seed>"));
	return newVector.elements();
    }

    public String[] getOptions()
    {
       String [] options = new String [2];
       int current = 0;
       options[current++] = "-S"; options[current++] = "" + this.m_seed;       
       return options;
    }  
    
    public String toString() 
    {
        StringBuffer msg = new StringBuffer();
        msg.append("SvmSimpleCluster Active Learner\n-----------------------");
        msg.append("The SVM used:\n" + this.m_svm.toString());
        return msg.toString();
    } 
    
    public static void main(String[] args) {
        try {
            System.out.println(ActiveEvaluationPlotter.evaluateModelAndPlot(new SvmMaxCEMInfo(), args));
            ActiveEvaluationPlotter.Plot(m_selfTestInfo, "entropy");                                           
        } catch (Exception e) {            
            System.err.println(e.getMessage());            
            e.printStackTrace();
        }  
    }        
}
