package weka.core;

import weka.core.*;
import java.util.*;
import java.io.*;

public class InstancesAscSet implements weka.core.DataRealizator
{
    private int m_numSets=0;
    private boolean m_sparseModeFlag=false;
    private int m_currTrainCache=-1;
    private int m_currTestCache=-1;
    private String m_dataSetPath;
    private String m_dataSetName;
    private Instances[] m_trainingSet;
    private Instances[] m_testingSet;
    private String m_fileType = null;    
    private int m_type = NO_TYPE;
    // for RND modes
    private Instances m_oneFoldTrainingSet;
    private Random m_rndm;
    private int m_numRndInst;    
    
    private static final int NO_TYPE = 0;
    private static final int ASC = -1;
    private static final int ARFF = -2;
    private static final int ASC_RND = -3;
    private static final int ARFF_RND = -4;
    private static final String m_trainDataSuffix = new String("_train_data_");
    private static final String m_trainLabelsSuffix = new String("_train_labels_");
    private static final String m_testDataSuffix = new String("_test_data_");
    private static final String m_testLabelsSuffix = new String("_test_labels_");
    private static final String m_sparse = new String("SPARSE");
    
    public InstancesAscSet(String dirName) throws Exception
    {
        int i;
        
        File dir = new File(dirName);
        if (!dir.isDirectory()) {
            throw new Exception("InstancesAscSet Ctor receives a directory");            
        }
        String[] files = dir.list();
        int slashIndx = dirName.lastIndexOf("/");
        this.m_dataSetName = new String(dirName.substring(slashIndx+1, dirName.length()));
        this.m_dataSetPath = new String(dirName + "/" + this.m_dataSetName);
        this.m_numSets=0;        
        this.m_fileType = null;
        for (i=0; i<files.length; i++) {
            if (this.m_fileType == null) {
                if (files[i].endsWith(".asc")) {
                    this.m_fileType = new String(".asc");
                    this.m_type = ASC;
                } else if (files[i].endsWith(".arff")) {
                    this.m_fileType = new String(".arff");
                    this.m_type = ARFF;
                }                
            }
            if ((this.m_fileType != null) && files[i].endsWith(m_fileType)) {
                this.m_numSets++;                            
            }
            if (files[i].compareTo(this.m_sparse) == 0) {
                this.m_sparseModeFlag = true;
            }            
        }
        if (this.m_fileType == null) throw new Exception("InstancesSet: unknown format");
        if (this.m_type == ASC) {
            this.m_numSets /= 4;
        } else if (this.m_type == ARFF) {
            this.m_numSets /= 2;
        } else {
            throw new Exception("InstancesSet: unknown format");
        }        
        this.m_trainingSet = new Instances[this.m_numSets];
        this.m_testingSet = new Instances[this.m_numSets];        
        for (i=0; i<this.m_numSets; i++) {
            this.m_trainingSet[i] = null;
            this.m_testingSet[i] = null;            
        }    
        this.m_currTrainCache=0;
        this.m_currTestCache=0;        
        this.trainCVNoCopy(this.m_numSets, 0);
        this.testCVNoCopy(this.m_numSets, 0);         
        // no RND mode for now
        this.m_oneFoldTrainingSet = null;
        this.m_rndm = null;
        this.m_numRndInst = -1;
    }
    
    private String buildFileName(String suffix, int indx)    
    {
        return new String(this.m_dataSetPath + suffix + (indx+1) + this.m_fileType);
    }
    
    public Instances trainCVNoCopy(int numFolds, int numFold) throws Exception
    {
        if (numFolds > this.m_numSets) throw new Exception("InstancesAscSet: number of folders to large");
        if (this.m_trainingSet[numFold] != null) {
            return this.m_trainingSet[numFold];
        }
        Instances instances;
        if (this.m_type == ASC) {
            instances = new Instances(this.buildFileName(m_trainDataSuffix, numFold),
                                      this.buildFileName(m_trainLabelsSuffix, numFold),
                                      this.m_dataSetName,
                                      this.m_sparseModeFlag);
            instances = clearEmptyInstances(instances);
        } else if (this.m_type == ARFF) {
	    BufferedReader reader = new BufferedReader(new FileReader(this.buildFileName(m_trainDataSuffix, numFold)));            
            instances = new Instances(reader);
            instances.setClassIndex(instances.numAttributes() - 1);
        } else if ((this.m_type == ASC_RND) || (this.m_type == ARFF_RND)) {
            int i, indx, size;            
            Instances copyInst;
            Instance ins;
            
            size = this.m_oneFoldTrainingSet.numInstances();
            copyInst = new Instances(this.m_oneFoldTrainingSet, 0);
            for (i=0; i<size; i++) {                
                copyInst.addNoCopy(this.m_oneFoldTrainingSet.instance(i));
            }
            instances = new Instances(this.m_oneFoldTrainingSet, 0);                        
                                    
            for (i=0; i<this.m_numRndInst; i++) { 
                indx = (int)(Math.ceil((copyInst.numInstances()-1) * this.m_rndm.nextDouble()));
                instances.addNoCopy(copyInst.instance(indx));
                copyInst.delete(indx);
            }                    
            //this.printSetToFileSvmLight(instances, "train.dat");            
        } else {
            throw new Exception("InstancesAscSet: unknown format");
        }
        this.m_trainingSet[numFold] = instances;
        if (this.m_currTrainCache != numFold) {            
            this.m_trainingSet[this.m_currTrainCache]=null;
            this.m_currTrainCache = numFold;
        }
        System.out.println("\nTrainFold " + numFold + "\n" + this.printSet(instances));        
        return instances;                                            
    }

    public Instances trainCV(int numFolds, int numFold) throws Exception
    {
        Instances instances = this.trainCVNoCopy(numFolds, numFold);
        return new Instances(instances);
    }

    public Instances trainCVRatio(int numFolds, int numFold, double ratio) throws Exception
    {
        return this.trainCVNoCopy(numFolds, numFold);
    }
    
    public Instances testCVNoCopy(int numFolds, int numFold) throws Exception
    {
        if (numFolds > this.m_numSets) throw new Exception("InstancesAscSet: number of folders to large");
        if ((this.m_type == ASC_RND) || (this.m_type == ARFF_RND)) numFold = 0;
        if (this.m_testingSet[numFold] != null) {
            return this.m_testingSet[numFold];
        }    
        Instances instances;
        if ((this.m_type == ASC) || (this.m_type == ASC_RND)) {
            instances = new Instances(this.buildFileName(m_testDataSuffix, numFold),
                                      this.buildFileName(m_testLabelsSuffix, numFold),
                                      this.m_dataSetName,
                                      this.m_sparseModeFlag); 
            instances = clearEmptyInstances(instances);            
        } else if ((this.m_type == ARFF) || (this.m_type == ARFF_RND)) {
	    BufferedReader reader = new BufferedReader(new FileReader(this.buildFileName(m_testDataSuffix, numFold)));            
            instances = new Instances(reader);
            instances.setClassIndex(instances.numAttributes() - 1);            
        } else {
            throw new Exception("InstancesSet: unknown format");
        }        
        this.m_testingSet[numFold] = instances;
        if (this.m_currTestCache != numFold) {
            this.m_testingSet[this.m_currTestCache] = null;
            this.m_currTestCache = numFold;
        }
        System.out.println("\nTestFold " + numFold + "\n" + this.printSet(instances));
        return instances;                                            
    }

    public Instances testCV(int numFolds, int numFold) throws Exception
    {        
        Instances instances = this.testCVNoCopy(numFolds, numFold);
        return new Instances(instances);
    }

    public Instances testCVRatio(int numFolds, int numFold, double ratio) throws Exception
    {
        return this.testCVNoCopy(numFolds, numFold);
    }
    
    private Instances clearEmptyInstances(Instances instances) throws Exception
    {
        Instance ins = instances.instance(0);
        if (!(ins instanceof SparseInstance)) return instances;
        int i, size = instances.numInstances();
        Instances copyInst = new Instances(instances, 0);        
        SparseInstance sIns;
        
        for (i=0; i<size; i++) {                
            sIns = (SparseInstance)(instances.instance(i));
            if (sIns.ids().size() > 1) {
                copyInst.addNoCopy((Instance)sIns);
            }
        }
        return copyInst;                                           
    }
    
    public void stratify(int numFolds) throws Exception
    {
        // nop
    }
    
    public void randomize(Random random) throws Exception
    {
        // nop
    }
    
    public void randomize(Random random, int numRndFolds, int numRndInst) throws Exception
    {
        if (numRndInst == -1) return;
        if (numRndInst <= 0) throw new Exception("InstancesAscSet - RndInst <= 0 " + numRndFolds);
        if (numRndFolds <= 0) throw new Exception("InstancesAscSet - RndFolds <= 0 " + numRndFolds);        
        if (this.m_numSets != 1) throw new Exception("InstancesAscSet - RndFolds are supported only when 1 original exist");        
        // this.m_numSets==1, numRndFolds>0
        int i;
        if (this.m_type == ASC) {
            this.m_type = ASC_RND;
        } else if (this.m_type == ARFF) {
            this.m_type = ARFF_RND;
        }
        if (this.m_trainingSet[0] == null) throw new Exception("Testing set is null");
        this.m_oneFoldTrainingSet = this.m_trainingSet[0];
        this.m_rndm = random;
        this.m_numRndInst = numRndInst;
        this.m_numSets = numRndFolds;
        this.m_trainingSet = new Instances[this.m_numSets];        
        for (i=0; i<this.m_numSets; i++) {
            this.m_trainingSet[i] = null;            
        }    
        this.m_currTrainCache=0;       
        this.trainCVNoCopy(this.m_numSets, 0);        
    }

    public int defaultNumFolds()
    {
        return this.m_numSets;
    }
    
    public int numAttributes()
    {
        return this.m_trainingSet[this.m_currTrainCache].numAttributes();
    }
    
    public int numInstances()
    {
        return this.m_trainingSet[this.m_currTrainCache].numInstances();
    }
    
    public Instances emptySet()
    {
        return new Instances(this.m_trainingSet[this.m_currTrainCache], 0);
    }    
    
    private String printSet(Instances instances) throws Exception
    {
        int posCount=0, negCount=0;
        Enumeration instEnum = instances.enumerateInstances();
        Instance ins;
        StringBuffer res = new StringBuffer();
        res.append("Relation " + instances.relationName() + "\n");
        while (instEnum.hasMoreElements()) {
            ins = (Instance)(instEnum.nextElement());
            if ((int)(ins.classValue()) == 0) {
                negCount++;                
            } else {
                posCount++;
            }
        }
        res.append("Total number of instances: " + (posCount+negCount) + "\n");
        res.append("Positive instances: " + posCount + "\n");
        res.append("Negative instances: " + negCount + "\n");
        res.append("Class index " + instances.classIndex() + "\n");
        return res.toString();
    }
    
    private void printSetToFileSvmLight(Instances instances, String fileName) throws Exception
    {
        int i, numInst = instances.numInstances();
        SparseInstance sIns;
        StringBuffer text = new StringBuffer();        
        FileOutputStream outFile = new FileOutputStream(fileName);                    
        for (i=0; i<numInst; i++) {            
            sIns = (SparseInstance)(instances.instance(i));          
            sIns.toStringBuffSvmLight(text);           
        }
        outFile.write(text.toString().getBytes());
        outFile.close();
    }
}